<?php

/*
SINX (SINX Is Not XML) is a lightweight semi-structured data format alternative
to XML, more convenient to use in ad hoc tasks. It provides same basic features
as XML does while being much simpler and less verbose. More information about
SINX is currently available at http://mikle33.narod.ru/sinx (in Russian).
Brief (but complete none the less) format description:
SINX level 0 - syntax. Document is a SINX string, which may consist of following
elements (or be empty):
- data run: sequence of any characters except for '<' and '>', means simply
that sequence of characters,
- verbatim run:
<delimiter%sequence of any characters%delimiter> 
(where delimiter is (possibly empty) sequence of any characters except for '<',
'>', '=', '%' and whitespaces, and second delimiter is exactly the same
sequence), means the sequence of characters starting just after post-delimiter
'%' and ending just before pre-closing delimiter '%'; it may contain any
characters as long as they don't form sequence of '%delimiter>', which marks end
of the sequence,
- special symbol:
<name data>
or
<name=data>
(where name is sequence of any characters except for '<', '>','=', '%' and
whitespaces, data is a SINX string; in first version, delimiter between name and
data may be arbitrarily long sequence of any whitespaces, including zero of them
if data starts with '<' or is empty), means a special symbol with given name and
data. The data itself is a SINX string too, and all above considerations are
recursively applicable to it. If the name-to-data delimiter is whitespaces, the
data starts at first non-whitespace character; if delimiter is '=', the data
starts immediately after '='. Special symbol is roughly an analogue of XML tag
<name>data</name>. Name may start with '.', which is called an attribute trait,
and/or end with ':', which is called a compound trait. If the delimiter between
name and data is '=', it is called a terminal trait. Symbol traits are not
included in the name (i. e., names 'sym.name', '.sym.name', '.sym.name:' and
'sym.name:' are syntactically identical), but their presence must be noted. They
have no predefined meaning at level 0.
All payload characters in data run, verbatim run and names are taken verbatim,
they are not case-converted, whitespaces are not merged/trimmed, etc. (Names are
therefore case-sensitive.)
If topmost SINX string has some syntax error, it may optionally be reported, but
the result must still be parsed using following options:
a) unclosed special symbol (<name ...EOF) either must be considered implicitly
closed at point of EOF, or its opening part (<name(whitespaces-or-'=') must be
ignored, which method of these two to choose is up to implementation (this one
uses first method),
b) unclosed verbatim run (<delimiter% ...EOF) must be considered implicitly
closed at point of EOF,
c) unmatched closing '>' (blahblah>...) must be either ignored or considered an
EOF mark, which method of these two to choose is up to implementation (this one
uses first method),
d) special symbol opening part exactly before EOF and with no delimiters
following (<nameEOF) is considered an opening part of special symbol with
whitespace name-to-data delimiter and must be handled as particular case of a).
SINX level 1 - basic metasyntax recommendations. Special symbols with some names
must be interpreted in special ways:
'--' (<-- blah>) - a comment, this special symbol must be entirely ignored,
'#dec-or-0[xX]Hex' (<#10>, <#0xA7>, <#0XdB>, etc.) - must be interpreted as
plain character with corresponding numerical code in implementation encoding,
data for such special symbols must be ignored; such characters are taken
verbatim as part of data runs they are inlined in,
'lt', 'gt' (case insensitive, i. e., <Lt>, <GT> etc. will also do) - must be
interpreted as plain characters '<' and '>' respectively, effectively the same
as <#60> and <#62> (for ASCII encoding).
Also, recommended (but not required) uses for special symbol traits are:
- terminal trait - for special symbols whose data is empty or plain string with
no special characters (except for possibly comments and encoded plain chars as
per directions above),
- attribute trait - for special symbols that represent properties of enclosing
special symbols, and no more than one symbol of such name is assumed to exist
within same parent (if there is, only the first or only the last one should be
in effect, depending on user format decision); use special symbols with
attribute trait for data that you would place in an XML attribute,
- compound trait - for data-only special symbols that are only assumed to hold
structured data (other special symbols) and no text; all data and verbatim runs
in such symbols (but not recursively) are assumed to be ignored.
*/

/// @brief SINX string
class SinxString
{
	const SINX_REGEXP =
		'/(?P<data_run>[^<>]+)|(?P<verbatim_run><(?P<verb_delim>[^\s=%<>]*)%(?P<verb_data>.*)(%(\k<verb_delim>)>|$))|(<(?P<spec_name>[^\s=%<>]*)(?P<spec_delim>\s*(?![=%])|=)(?P<spec_data>(?R)*)(>|$))/';
	private $components = null;
	private $string;

	/// @brief Construct from a string by parsing it according to SINX level 1
	/// @param The string
	public function __construct ($string='')
	{
		if (!is_string ($string))
			trigger_error ('SinxString constructor must take a string. For other options see SinxString::from_components',E_USER_ERROR);
		$this->string = $string;
	}

	/// @brief Get SINX string components, all or special symbols only
	/// @param $specials If true, only special symbol entries will be included
	/// @return array, where each element is either string or SinxSpecial
	public function components ($specials_only=false)
	{
		if (!is_null ($this->components))
			return $this->components;
		else
		{
			// parse (in lazy manner)
			preg_match_all (SinxString::SINX_REGEXP,$this->string,$matches,PREG_SET_ORDER);
			$this->string = null;
			$was_string = false;
			$was_spec = false;
			$this->components = array ();
			foreach ($matches as &$match)
			{
				if (isset ($match['spec_delim']) || isset ($match['spec_data']))
				{
					// a special char:
					// check for comment and char codes
					$name = trim ($match['spec_name'],'.:');
					if ($name=='--') continue; // comment, ignore
					if (preg_match ('/^(#0[xX][0-9A-Fa-f]+|#[0-9]+|[lLgG][tT])$/',$name,$as_num)>0)
					{
						// char code
						$as_num = str_replace (
							array ('lt','lT','Lt','LT','gt','gT','Gt','GT'),
							array ('#60','#60','#60','#60','#62','#62','#62','#62'),
							$as_num[1]);
						sscanf ($as_num,"#%i",$chr);
						$chr = chr ($chr);
						if ($was_string)
							$this->components[sizeof ($this->components)-1] .= $chr;
						else
							$this->components[] = $chr;
						$was_spec = false;
						$was_string = true;
						continue;
					}
					$this->components[] = new SinxSpecial
						($match['spec_name'],$match['spec_data'],
						$match['spec_delim']=='=');
					$was_spec = true;
					$was_string = false;
				}
				else
				{
					// string
					$data = isset ($match['verb_data'])?
						$match['verb_data'] : $match['data_run'];
					if ($was_string)
						$this->components[sizeof ($this->components)-1] .= $data;
					else
						$this->components[] = $data;
					$was_string = true;
					$was_spec = false;
				}
			}
			if (!function_exists ('_is_component_special'))
			{
				function _is_component_special (&$spec)
				{ return !is_string ($spec); }
			}
			return $specials_only? array_merge
				(array_filter ($this->components,'_is_component_special'),array ())
				: $this->components;
		}
	}

	/// @brief Extract all special symbols grouped by name
	/// @return array: array['spec_name'][index] => SinxSpecial
	public function specials ()
	{
		$result = array ();
		$comps = $this->components ();
		foreach ($comps as &$comp)
		if (is_object ($comp) && get_class ($comp)=='SinxSpecial')
			$result[$comp->get_name ()][] = $comp;
		return $result;
	}

	/// @brief Return $index-th (0-based) of special symbols with given name.
	/// NULL if there is no such symbol.
	public function special ($spec_name,$index=0)
	{
		$comps = $this->specials ();
		return @$comps[$spec_name][$index];
	}

	/// @brief Construct SINX string from list of components (complement to
	/// components ())
	/// @param $comps Array of components, each element must be either
	/// SinxSpecial or type convertible to string
	public static function from_components (array $comps)
	{
		$result = new SinxString ();
		$result->string = null;
		$result->components = array ();

		$was_string = false;
		foreach ($comps as &$comp)
		if (is_object ($comp) && get_class ($comp)=='SinxSpecial')
		{
			$was_string = false;
			$result->components[] = $comp;
		}
		else
		{
			// merge subsequent string fragments
			if ($was_string)
				$result->components[sizeof ($result->components)-1] .=
					strval ($comp);
			else
				$result->components[] = strval ($comp);
			$was_string = true;
		}
		return $result;
	}

	/// @brief Return string representation of the SINX string
	/// @return The string representation
	public function as_string ()
	{
		if (!is_null ($this->string)) return $this->string;
		else
		{
			$frags = array ();
			foreach ($this->components () as $comp)
			$frags[] = is_string ($comp)
				? SinxString::quote ($comp,(!sizeof ($frags))? '<' : '')
				: sprintf ("<%s%s%s%s%s>",
					$comp->is_attr ()? '.' : '',
					$comp->get_name (),
					$comp->is_compound ()? ':' : '',
					$comp->is_terminal ()? '=' :
						(sizeof ($comp->get_data ()->components ())==0 ? '' : ' '),
					$comp->get_data ()->as_string ()
				);
			return implode ('',$frags);
		}
	}

	/// @brief Quotes SINX meta characters (namely, '<' and '>') in the string
	/// according to SINX level 1 (converts them to '<#60>' and '<#62>').
	/// Optionally, you can apply escape to any other characters, specify them
	/// in $extra_chars.
	/// @param $string String to escape
	/// @param $extra_chars Set of additional chars (besides '<' and '>') to be
	/// escaped. They are converted to '<#code>' specials.
	/// '<' char in $extra_chars denotes a leading whitespace, useful when you
	/// need to outline the meaningfulness of leading whitespaces in the string.
	/// @return Escaped string
	public static function quote ($string,$extra_chars='')
	{
		$extra_chars = str_replace ('<','$\s',preg_quote ($extra_chars,'/'));
		$preg = "/([<>{$extra_chars}])/";
		$frags = preg_split ($preg,$string,-1,PREG_SPLIT_DELIM_CAPTURE);
		for ($i=1; $i<sizeof ($frags); $i+=2)
			$frags[$i] = '<#'.ord ($frags[$i]).'>';
		return implode ($frags,'');
	}
}

/// @brief SINX special symbol
class SinxSpecial
{
	private $name;
	private $data;
	private $is_attr,$is_compound,$is_terminal;

	public function is_attr () { return $this->is_attr; }
	public function is_compound () { return $this->is_compound; }
	public function is_terminal () { return $this->is_terminal; }

	/// @brief Set name. Starting dot and trailing colon (attr & compound
	/// traits) are properly recognized and removed from the name itself.
	public function set_name ($name)
	{
		if (!is_string ($name))
			// string
			trigger_error ('SinxSpecial::set_name: name must be a string',E_USER_ERROR);
		else if (strcspn ($name,"<> \n\r\t%=")<strlen ($name))
			// invalid character in the name found
			trigger_error ("SinxSpecial::set_name: name '{$name}' contains a disallowed character",E_USER_ERROR);
		else
		{
			if (strlen ($name)>0)
			{
				$this->is_attr = $name{0}=='.';
				$this->is_compound = $name{strlen ($name)-1}==':';
				$name = trim ($name,'.:');
			}
			else
				$this->is_attr = $this->is_compound = false;
			$this->name = $name;
			return;
		}
		// default for an invalid name
		$this->name = '';
	}

	public function get_name ()
	{
		return $this->name;
	}

	/// @brief Set spec symbol data
	/// @param $data the data to set, as SinxString or a string representation
	public function set_data ($data,$terminal_trait=false)
	{
		if (is_object ($data) && get_class ($data)=='SinxString')
			$this->data = $data;
		else
			$this->data = new SinxString (strval ($data));
		$this->is_terminal = $terminal_trait;
	}

	/// @brief Return spec symbol data
	/// @return Spec symbol data, as SinxString
	public function get_data ()
	{
		return $this->data;
	}

	public function __construct ($name,$data='',$terminal_trait=false)
	{
		$this->set_name ($name);
		$this->set_data ($data);
		$this->is_terminal = $terminal_trait;
	}
}

//
// some standalone helper functions:
//

/// @brief Import a piece of structured data from a SINX string using suggested
/// blueprint.
///
/// Data imported from SINX is arranged into an associative or int-indexed array
/// with each string-indexed element imported from SINX symbol with
/// corresponding name, each element can be int, bool, string, float,
/// int-indexed or associative array of same types of elements, or a SinxString
/// type object. SINX does not provide redundant meta information, so you must
/// specify structure expected from the SINX string explicitly, using a schema
/// called blueprint. A blueprint is an associative array that describes layout
/// of resulting array and specifies defaults, its structure is best explained
/// by example:
/// array (
///  // resulting array will contain string key "A", which must be imported
///  // from spec symbol named "A" at root level. "x" denotes type of expected
///  // value (string in this case), and also specifies its default value that
///  // will be set if spec symbol A doesn't exist.
///  // If there are more than one A at this level, only first one will be
///  // considered.
///  "A" => "x",
///  "B" => 10, // same for int and key B
///  "C" => true, // same for bool
///  "D" => 1.1, // same for float
///  // specifies an array of entries, which are specified by multiple spec
///  // symbols named "E", one symbol per entry, in same order as they are in
///  // a given SINX string. If no such symbols, this entry in the result
///  // will be an empty array.
///  // Each subsymbol is specified by blueprint array ( "F" => 11 ).
///  "E" => array (
///   array ( // first member of the array specifies blueprint for each array entry
///    // which works the same way at level of array entry's character
///    "F" => 11
///   )
///  ),
///  // specifies an array of entries, which are specified by multiple spec
///  // symbols named "E1", one symbol per entry, in same order as they are in
///  // a given SINX string. If no such symbols, this entry in the result
///  // will be an empty array.
///  // Each subsymbol is assumed to hold value representing int (typeof (100)),
///  // (specify bool, string, float and SinxString similarly, by providing
///  // a value of this type as 1st and only unkeyed entry). The value itself
///  // is not used in this case, it just denotes the type.
///  "E1" => array ( 100 ),
///  // data of corresponding symbol ("G") will be presented as unconverted
///  // SinxString that you can manipulate manually, default is empty SinxString
///  "G" => new SinxString (),
///  // expects single special symbol named "H" at root level, takes its data
///  // and imports into this entry as an associative array, using the value
///  // as blueprint to apply to H's data for recursive sinx_import
///  "H" => array (
///   "I" => true
///  )
///  // expect a single symbol named "J" with no particular content, default
///  // is empty array (string keys may occur)
///  "J" => array (),
///  // expect multiple symbols named "K" with no particular content, arrange
///  // them into numeric indexed array
///  "K" => array (array ())
/// );
/// This blueprint could match such SINX string as:
/// <A x1><B 100><C on><D 2><E <F 500>><E <F 12>><G smth><H <I off>><J><K><K><K>
/// and import using it would yield:
/// array (
///  "A" => "x1", "B" => 100, "C" => true, "D" => 2.0,
///  "E" => array (array ("F" => 500), array ("F" => 12)),
///  "E1" => array (), // if there were entries like <E1 1><E1 2>, it'd be
///   // "E1" => array (1,2) (of ints as 100 in "E1" => array (100) of the
///   // blueprint is int)
///  "G" => new SinxString ("smth"),
///  "H" => array ("I" => false),
///  "J" => array (), // (string keys may occur)
///  "K" => array (array (),array (),array ())
/// );
/// Note: bools are imported as true for values 'on', 'yes', 'true' and
/// 'enabled', and as false for all other values.
/// @param $blueprint Description of the blueprint.
/// @param $sinx_string SINX string to import from, string or SinxString 
/// @param $store_unspecified If true, special symbols with names that have no
/// corresponding string keys in the blueprint will be put in resulted array as
/// SinxString-s as if they were specified with "name" => new SinxString (). If
/// false, such symbols will be ignored (i. e., result will only contain fields
/// explicitly specified in the blueprint, and nothing besides this.
/// @return Imported array
function sinx_import (array $blueprint,$sinx_string,$store_unspecified=true)
{
	function _sinx_init_by_blueprint (&$dest,&$blueprint)
	{
		foreach ($blueprint as $k=>&$v)
			if (is_array ($v))
				if (!isset ($v[0]))
					_sinx_init_by_blueprint ($dest[$k],$v);
				else
					$dest[$k] = array ();
			else
				$dest[$k] = $v;
	}
	function _do_sinx_import ($name,&$dest,&$blueprint,$sinx_string,$store_unspecified)
	{
		if (is_array ($blueprint))
		{
			$specs = $sinx_string->specials ();
			if (isset ($blueprint[0]))
			{
				// array of $name-s
				//if (!is_array ($blueprint[0]) || sizeof ($blueprint)!=1)
					//trigger_error ('sinx_import: blueprint array can only contain string keys or only a single unkeyed subblueprint array',E_USER_ERROR);
				if (sizeof ($blueprint)!=1)
					trigger_error ('sinx_import: subblueprint for numeric-indexed array can only contain a single unkeyed value',E_USER_ERROR);
				if (is_array ($blueprint[0]))
					// a subsymbol subblueprint
					_do_sinx_import (
						sizeof ($dest[$name]),
						$dest[$name],
						$blueprint[0],
						$sinx_string,
						$store_unspecified);
				else
				{
					// an integral value subblueprint
					$dest[$name][] = $blueprint[0];
					_do_sinx_import (
						sizeof ($dest[$name])-1,
						$dest[$name],
						$blueprint[0],
						$sinx_string,
						$store_unspecified);
				}
			}
			else
			{
				// a subsymbol
				$ss_dest = &$dest[$name];
				_sinx_init_by_blueprint ($ss_dest,$blueprint);
				foreach ($specs as $spec_name=>&$spec)
				{
					if (isset ($blueprint[$spec_name]))
						// key exists in the blueprint,
						// suggested value is $blueprint[$spec_name]
						foreach ($spec as &$sspec)
						_do_sinx_import (
							$spec_name,
							$ss_dest,
							$blueprint[$spec_name],
							$sspec,
							$store_unspecified);
					else
						// key doesn't exist, import as SinxString
						// (if store_unspecified is set)
						if ($store_unspecified)
						_do_sinx_import (
							$spec_name,
							$ss_dest,
							new SinxString (),
							$spec[0],
							$store_unspecified);
				}
			}
		}
		else
		{
			// a terminal node
			if (is_float ($dest[$name]))
				list ($dest[$name]) =
					sscanf ($sinx_string->as_string (),'%f');
			else if (is_bool ($dest[$name]))
				$dest[$name] =
					!strcasecmp ($sinx_string->as_string (),'yes') ||
					!strcasecmp ($sinx_string->as_string (),'on') ||
					!strcasecmp ($sinx_string->as_string (),'true') ||
					!strcasecmp ($sinx_string->as_string (),'enabled');
			else if (is_int ($dest[$name]))
				list ($dest[$name]) =
					sscanf ($sinx_string->as_string (),'%i');
			else if (is_string ($dest[$name]))
				$dest[$name] = $sinx_string->as_string ();
			else if ((!isset($dest[$name]) && $store_unspecified) ||
				(is_object ($dest[$name]) && get_class ($dest[$name])=='SinxString'))
				$dest[$name] = $sinx_string;
		}
	}
	if (!(is_object ($sinx_string) && get_class ($sinx_string)=='SinxString'))
		$sinx_string = new SinxString (strval ($sinx_string));
	$result = array ();
	$specs = $sinx_string->specials ();
	_sinx_init_by_blueprint ($result,$blueprint);
	foreach ($specs as $spec_name=>&$spec)
	{
		if (isset ($blueprint[$spec_name]))
			// key exists in the blueprint,
			// suggested value is $blueprint[$spec_name]
			foreach ($spec as &$sspec)
			_do_sinx_import (
				$spec_name,
				$result,
				$blueprint[$spec_name],
				$sspec,
				$store_unspecified);
		else
			// key doesn't exist, import as SinxString
			// (if store_unspecified is set)
			if ($store_unspecified)
			_do_sinx_import (
				$spec_name,
				$result,
				new SinxString (),
				$spec[0],
				$store_unspecified);
	}
	return $result;
}

/// @brief Export an associative or numeric-indexed array into a SINX string.
/// Array must be strictly associative, and its entries must be string, float,
/// int, bool, SinxString objects or arrays which are either strictly
/// associative (array ("a" => anything, "b" => anything)) or strictly numeric
/// indexed and comprise values of same type (array (1,2) or array ("a","b")
/// etc.).
/// @param $arr Array to export, see description for requirements
/// @return SinxString with exported data
function sinx_export (array $arr)
{
	$comps = array ();
	if (!function_exists ('do__sinx_export'))
	{
	function do__sinx_export (array &$comps,$key,array &$arr)
	{
		$num = $ass = false;
		foreach ($arr as $k=>&$v)
		{
			$num |= is_int ($k);
			$ass |= is_string ($k);
			if ($num && $ass)
				trigger_error ('sinx_export: array must be either strictly num-indexed or strictly associative',E_USER_ERROR);
			if (!is_array ($v))
			{
				// not an array entry
				$strval = strval ($v);
				if (is_bool ($v)) $strval = $v? 'on' : 'off';
				if (is_object ($v) && get_class ($v)=='SinxString')
					$strval = $v;
				if (is_int ($k))
					$comps[] = new SinxSpecial ($key,$strval);
				else
					$comps[] = new SinxSpecial ($k,$strval);
			}
			else if (isset ($v[0]) || sizeof ($v)<=0)
			{
				// a numerical-indexed (or empty) array entry - serialize here
				do__sinx_export ($comps,$k,$v);
			}
			else
			{
				// an associative array entry - serialize to SinxSpecial
				$comps2 = array ();
				do__sinx_export ($comps2,$k,$v);
				$comps[] = new SinxSpecial (
					is_int ($k)? $key : $k,SinxString::from_components ($comps2));
			}
		}
	}
	}
	do__sinx_export ($comps,0,$arr);
	return SinxString::from_components ($comps);
}

?>